;***********************************************************************
;
;	Microchip 16-bit Embedded Control Design Contest
;
;	Entry # MT2268
;
;	Spa Pump Controller - Spa Interface Processor
;
;***********************************************************************
;
;	Keypress Decoding and Generating Routines
;
;***********************************************************************


#include	<P16f687.inc>
#include	"SpaCtl.inc"


;***********************************************************************
;
;	Variables
;
;***********************************************************************

VARS		UDATA

		global	Key_timer
Key_timer	res	1		; Key scan interval timer
Key_debcnt	res	1		; Key press debounce counter
Key_holdcnt	res	1		; Key held down counter
Key_temp	res	1		; Temporary key value
Key_last	res	1		; Last incoming key value
		global	Key_val
Key_val		res	1		; Debounced key value
		global	Key_flag
Key_flag	res	1		; Key press flag


		global	Keyout_timer
Keyout_timer	res	1		; Outgoing key press timer
Keyout_val	res	1		; Outgoing key value


PROG	CODE


;***********************************************************************
;
;	Incoming key decoding subroutines
;
;***********************************************************************

; Key input handler called by main loop
;
;	This routine decodes and debounces incoming keypresses
;

	global	DoKeyIn
DoKeyIn
	movf	Key_timer,w	; Key scan interval elapsed?
	SKIPZ
	return			;  no

	movlw	KEY_TIME
	movwf	Key_timer	; Restart key scan timer

; Check if any key is pressed

	SETBANK	2
	movf	CM1CON0,w	; Get comparator 1 status
	SETBANK	0
	andlw	1<<C1OUT	; Any key pressed?
	BEQ	NoKeyIn		;  no, skip ADC conversion

; Decode the key value

	movlw	ADC_keypad
	CALLX	GetADC		; Get keypad voltage level
	addlw	0x10
	movwf	Key_temp
	rrf	Key_temp,f	; Shift carry into high bit
	swapf	Key_temp,w	; Swap nibbles
	andlw	0x0F		; Mask lower nibble to get key value
NoKeyIn
	movwf	Key_temp	; 1 - 8 = key value, 0 = no key pressed

; Debounce the key value

	xorwf	Key_last,w	; Key value same as before?
	BEQ	KeyOk		;  yes

	xorwf	Key_last,f	;  no, save new key value
	movlw	KEY_DEB
	movwf	Key_debcnt	;  and reinit debounce counter
	return

KeyOk
	movf	Key_debcnt,w	; Debounce counter = 0?
	BEQ	KeyDeb		;  yes, debounce already completed

	decfsz	Key_debcnt,f	; Decrement debounce counter
	return			; Not done yet

; Now we have a newly debounced keypress

	movf	Key_last,w
	BEQ	KeyRel		; Nothing pressed
	movwf	Key_val		; Save the new key value
	movwf	Key_flag	; Set the key pressed flag
	movlw	KEY_HOLD
	movwf	Key_holdcnt	; Init the key held down counter
	return

KeyRel
;;	movf	Key_val,w	; Get previous key
;;	andlw	0x0F
;;	iorlw	KV_rel
	movwf	Key_val
;;	movwf	Key_flag	; Set the key released flag
	return

; Do key autorepeat processing

KeyDeb
	movf	Key_val,w	; Any key pressed?
	BEQ	KeyDone		;  no
	decfsz	Key_holdcnt,f	; Decrement key hold counter
	return			; Not done yet

	btfsc	Key_val,K_HOLD	; Was key already held down before?
	bsf	Key_val,K_RPT	;  yes, set repeat flag

	bsf	Key_val,K_HOLD	; Set key hold flag

	movf	Key_val,w
	movwf	Key_flag	; Save the new key value
	movlw	KEY_RPT
	movwf	Key_holdcnt	; Init the key autorepeat counter
KeyDone
	return


;***********************************************************************
;
;	Outgoing key generating subroutines
;
;***********************************************************************

;  Send an outgoing keypress in W register to the spa control pack

	global	SendKey
SendKey
	andlw	0x0F
	xorwf	Keyout_val,w	; Same value as before?
	BEQ	SendSame	;  yes, just restart timer
	xorwf	Keyout_val,f	; Save new value
	xorwf	Keyout_val,w	; Get back old value
	SKIPZ			; Was it zero?
	call	PotClr		;  no, clear the pot first

	movf	Keyout_val,w	; Get new keypress value
	CALLX	KeyoutTbl	; Lookup pot count value
	call	PotSet		; Start sending the keypress

SendSame
	movlw	KEYOUT_TIME
	movwf	Keyout_timer	; Init time to send
	return

;
; Key output handler called by main loop
;
;	Check for end of outgoing keypress
;
	global	DoKeyOut
DoKeyOut
	movf	Keyout_val,w	; Outputting a key?
	SKIPNZ
	return			;  no, nothing to do

	movf	Keyout_timer,w	; Key output timer elapsed?
	SKIPZ
	return			;  no

	call	PotClr		; Stop sending the keypress

	clrf	Keyout_val	; Clear key value
	return



;***********************************************************************
;
;	Digital Pot control subroutines
;
;***********************************************************************

;  Set pot to value in W (0 - 63)

	global	PotSet
PotSet
	bsf	POT_PORT,POT_UD
	bcf	POT_PORT,POT_CS
	nop			; Need 3 usec from CS \_ to UD _/
	nop
PotSLp
	bcf	POT_PORT,POT_UD
	bsf	POT_PORT,POT_UD	; Toggle up one step
	addlw	-1		; Repeat?
	BNE	PotSLp		;  yes
	bsf	POT_PORT,POT_CS
	return


;  Clear pot to 0

	global	PotClr
PotClr
	movlw	64		; Make sure we go all the way down to 0
	bcf	POT_PORT,POT_UD
	bcf	POT_PORT,POT_CS
	nop			; Need 3 usec from CS \_ to UD _/
	nop
	nop
PotCLp
	bsf	POT_PORT,POT_UD	; Toggle down one step
	bcf	POT_PORT,POT_UD
	addlw	-1		; Repeat?
	BNE	PotCLp		;  yes
	bsf	POT_PORT,POT_UD
	bsf	POT_PORT,POT_CS
	return


;***********************************************************************
;
;	Initialization
;
;***********************************************************************

;  Initialize all key vars

	global	KeyInit
KeyInit
	clrf	Key_timer	; Init key debounce vars
	movlw	KEY_DEB
	movwf	Key_debcnt
	clrf	Key_last
	clrf	Key_val
	clrf	Key_flag

	clrf	Keyout_timer	; Init key output vars
	clrf	Keyout_val
	return

	end
