;***********************************************************************
;
;	Microchip 16-bit Embedded Control Design Contest
;
;	Entry # MT2268
;
;	Spa Pump Controller - Spa Interface Processor
;
;***********************************************************************
;
;	LCD Control Routines
;
;***********************************************************************


#include	<P16f687.inc>
#include	"SpaCtl.inc"


;***********************************************************************
;
;	Variables
;
;***********************************************************************

	extern	Sys_fault


VARS		UDATA

LCD_size	equ	13		; Size of LCD segment data block (bytes)

LCD_buf		res	LCD_size	; LCD segment data

LCD_ptr		res	1		; LCD buffer pointer

		global	LCD_flag
LCD_flag	res	1		; LCD buffer full flag

		global	LCD_icons
LCD_icons	res	1		; LCD status icon flags
					; (see bit definitions in SpaCtl.inc)

		global	LCD_timer
LCD_timer	res	1		; LCD frame timer

LCD_TIME	equ	20		; Maximum time between frames (msec)


INT_VARS	UDATA_SHR
	   
w_save		res	1		; Put interrupt vars in shared area
status_save	res	1
fsr_save	res	1


PROG	CODE


;***********************************************************************
;
;	LCD subroutines
;
;***********************************************************************

;
; LCD handler called by main loop
;
;	This routine monitors the incoming LCD segment buffer,
;	scanning it for various icon on/off states as needed.
;	It also modifies the buffer before sending it out to
;	the display, if we want to add to the display content.
;
	global	DoLCD
DoLCD
	btfsc	LCD_flag,0	; LCD buffer full?
	goto	DoLCD1		;  yes

; Check if the frame timer has timed out, since this
; means the spa control pack must have gone away.
; The timer is reset each time a new frame is received.

	movf	LCD_timer,w	; LCD frame timer expired?
	SKIPZ
	return			;  no

	bsf	Sys_fault,FAULT_lcd	; Set frame timeout flag
	return

DoLCD1
	bcf	LCD_flag,0	; Clear buffer full flag

	movlw	LCD_TIME
	movwf	LCD_timer	; Restart LCD frame timer

; Check for display data modifications

	extern	State
	movf	State,w
	xorlw	4		; Are we in state 4 (speed adjust mode)?
	BNE	DoLCD2
	bcf	LCD_buf+6,2	;  yes, turn on clock icon
DoLCD2

; Check for displayed icon states

	clrf	LCD_icons	; Clear all flags
	btfss	LCD_buf+4,6	; Pump 1 icon on?
	bsf	LCD_icons,ICON_pump1	;  yes, set pump 1 flag

;;	goto	LCDprep

; Write the first byte to the SSP buffer,
; to prepare it for a slave mode transfer

	global	LCDprep
LCDprep
	movf	LCD_buf,w	; Get first byte from buffer
	movwf	SSPBUF		;  and write it to SSP
	return


;***********************************************************************
;
;	Initialization
;
;***********************************************************************

	global	LCDinit
LCDinit
	movlw	LCD_buf
	movwf	LCD_ptr		; Init buffer pointer
	clrf	LCD_flag	; Clear buffer full flag
	movlw	255
	movwf	LCD_timer	; Start LCD frame timer
	return


;***********************************************************************
;
;	Interrupt routine
;
;***********************************************************************

;
; The interrupt routine handles the transfer of LCD data
; between the SSP port and the LCD segment buffer.
; It sends the current buffer out, while overwriting
; it with new incoming data. Once the entire buffer
; has been transferred, it sets the buffer full flag.
; This lets the higher level routines know they can
; make modifications to the data before it's sent out
; the next time around.
;
; The higher level routines must call LCDprep to write
; out the first byte before the interrupt is enabled,
; and each time before a new buffer can be sent out.
;

	global	Interrupt
Interrupt

	movwf	w_save		; save off current W register contents
	swapf	STATUS,w	; swap STATUS register into W register
	movwf	status_save	; save off contents of STATUS register
	movf	FSR,w		; move FSR register into W register
	movwf	fsr_save	; save off contents of FSR register

	banksel	0

	bcf	PIR1,SSPIF	; Clear interrupt flag

	movf	LCD_ptr,w	; Get LCD buffer pointer
	movwf	FSR
	movf	SSPBUF,w	; Get a byte from SPI port
	movwf	INDF		;  and save it in the buffer

	incf	LCD_ptr,f	; Increment the pointer
	movf	LCD_ptr,w
	sublw	LCD_buf+LCD_size; At end of buffer?
	BEQ	LCDFull		;  yes, all done for now

	incf	FSR,f
	movf	INDF,w		; Get a byte from buffer
	movwf	SSPBUF		;  and prepare it for SPI transmission
	goto	IntDone
LCDFull
	movlw	LCD_buf
	movwf	LCD_ptr		; Re-init buffer pointer
	bsf	LCD_flag,0	;  and set flag for LCD processing
IntDone
	movf	fsr_save,w	; retrieve copy of FSR register
	movwf	FSR		; restore pre-isr FSR register contents
	swapf	status_save,w	; retrieve swapped copy of STATUS register
	movwf	STATUS		; restore pre-isr STATUS register contents
	swapf	w_save,f
	swapf	w_save,w	; restore pre-isr W register contents
	retfie			; return from interrupt

	end
