;***********************************************************************
;
;	Microchip 16-bit Embedded Control Design Contest
;
;	Entry # MT2268
;
;	Spa Pump Controller - Spa Interface Processor
;
;***********************************************************************
;
;	Pump Drive Unit Serial Communication Routines
;
;***********************************************************************


#include	<P16f687.inc>
#include	"SpaCtl.inc"


;***********************************************************************
;
;	Variables
;
;***********************************************************************

	extern	Sys_fault


VARS		UDATA

SER_SIZE	equ	3		; Size of serial data block (bytes)

Ser_in		res	SER_SIZE	; Serial receive buffer
Ser_out		res	SER_SIZE	; Serial transmit buffer
Ser_in_cnt	res	1		; Serial receive count
Ser_out_cnt	res	1		; Serial transmit count


Pump1_freq	res	1		; Pump 1 operating frequency
		global	Resp_timer
Resp_timer	res	1		; Command response timer

Resp_flag	res	1		; Response flag:

RESP_WAIT	equ	0		; Waiting for response flag

		global	Ping_timer
Ping_timer	res	1		; Ping timer

PING_TIME	equ	250		; Ping interval (msec)


PROG	CODE


;***********************************************************************
;
;	Pump drive unit communication subroutines
;
;***********************************************************************
;
; Pump control handler called by main loop
;
;	Check to see if we need to send out a ping
;	to keep the drive unit from thinking we've crashed
;
	global	DoPump
DoPump
	movf	Ping_timer,w	; Ping interval elapsed?
	SKIPZ
	return			; no

	goto	Pump1Send	; Send a command


; Set the pump to the initial frequency setting

	global	Pump1Init
Pump1Init
	bcf	Sys_fault,FAULT_resp	; Clear response timeout flag

	movlw	FREQ_INIT*4	; Convert initial freq to Hz * 4
	movwf	Pump1_freq
	goto	Pump1Send	;  and send it

; Increase the pump speed

	global	Pump1Up
Pump1Up
	movf	Pump1_freq,w	; Get current frequency
	sublw	FREQ_MAX*4	; At upper limit?
	SKIPC
	return			;  yes
	incf	Pump1_freq,f	; Increment it
	goto	Pump1Send	;  and send it

; Decrease the pump speed

	global	Pump1Down
Pump1Down
	movf	Pump1_freq,w	; Get current frequency
	sublw	FREQ_MIN*4+1	; At lower limit?
	SKIPNC
	return			;  yes
	decf	Pump1_freq,f	; Decrement it
	goto	Pump1Send	;  and send it

; Turn pump off

	global	Pump1Off
Pump1Off
	clrf	Pump1_freq	; Set frequency to 0
;;	goto	Pump1Send	;  and send it


; Send the Set Frequency command to the pump drive unit

Pump1Send
	movlw	CMD_set_spd	; Set Frequency command
	movwf	Ser_out+0	; Save it in transmit buffer
	movlw	Pump1_freq	; Get current frequency
	movwf	Ser_out+1	;  and save it in buffer
	addlw	CMD_set_spd
	xorlw	0xFF		; Compute checksum
	movwf	Ser_out+2	;  and save it in buffer
	movlw	SER_SIZE
	movwf	Ser_out_cnt	; Set transmit byte count

	movlw	RESP_TIME
	movwf	Resp_timer	; Start command response timer
	movlw	1<<RESP_WAIT
	movwf	Resp_flag	; Set waiting for response flag
	movlw	PING_TIME
	movwf	Ping_timer	; Reset ping timer
	return


;***********************************************************************
;
;	Serial I/O subroutines
;
;***********************************************************************

;
; Serial handler called by main loop
;
;	This routine processes incoming serial data,
;	and transmits any data waiting to be sent.
;

	global	DoSerial
DoSerial

; Capture a byte and save it in the receive buffer

	btfss	PIR1,RCIF		; Byte received?
	goto	SerNoRx			;  no
	movf	Ser_in_cnt,w
	addlw	Ser_in			; Compute buffer pointer
	movwf	FSR
	movf	RCREG,w			; Read the byte
	btfss	Resp_flag,RESP_WAIT	; Waiting for a response?
	goto	SerNoRx			;  no, ignore the byte
	movwf	INDF			; Save it in buffer
	incf	Ser_in_cnt,f		; Update byte count
	movf	Ser_in_cnt,w
	xorlw	Ser_in+SER_SIZE		; Have we filled up the buffer?
	BNE	SerNoRx			;  no

; Process the incoming data

	movf	Ser_in+0,w
	addwf	Ser_in+1,w
	addwf	Ser_in+2,w		; Compute checksum
	xorlw	0xFF			; Valid?
	BNE	SerRxErr		;  no, ignore data
	movf	Ser_in+0,w		; Get command byte
	xorwf	Ser_out+0,w		; Same as sent out earlier?
	BNE	SerRxErr		;  no, ignore data

	clrf	Resp_flag		; Valid response, clear flag
SerRxErr
	clrf	Ser_in_cnt		; Clear input buffer
SerNoRx

; Process any outgoing data

	movf	Ser_out_cnt,w		; Any data waiting to be sent?
	BEQ	SerNoTx			;  no
	btfss	PIR1,TXIF		; TX buffer empty?
	goto	SerNoTx			;  no
	sublw	Ser_out+SER_SIZE	; Compute buffer pointer
	movwf	FSR
	movf	INDF,w			; Get a byte
	movwf	TXREG			;  and transmit it
	decf	Ser_out_cnt,f		; Decrement byte count
SerNoTx

; Check for timeout if we're waiting for a response

	btfss	Resp_flag,RESP_WAIT	; Waiting for response?
	goto	SerNoResp		;  no
	movf	Resp_timer,w		; Timeout?
	BNE	SerNoResp		;  no
	bcf	Resp_flag,RESP_WAIT	; Clear waiting flag
	bsf	Sys_fault,FAULT_resp	;  and set timeout flag
	clrf	Ser_in_cnt		; Clear buffer count
SerNoResp

	return


;***********************************************************************
;
;	Initialization
;
;***********************************************************************

	global	SerInit
SerInit
	clrf	Ser_in_cnt	; Clear buffer counts
	clrf	Ser_out_cnt
	clrf	Resp_flag	; Clear response flag
	clrf	Ping_timer	; Clear ping timer
	return

	end
