;***********************************************************************
;
;	Microchip 16-bit Embedded Control Design Contest
;
;	Entry # MT2268
;
;	Spa Pump Controller - Spa Interface Processor
;
;***********************************************************************
;
;	Main Routine
;
;***********************************************************************


#include	<P16f687.inc>
#include	"SpaCtl.inc"


;***********************************************************************
;
;	Config Word
;
;***********************************************************************


	__config  _FCMEN_OFF & _IESO_OFF & _BOR_NSLEEP &_PWRTE_ON & _WDT_ON & _INTOSCIO


;***********************************************************************
;
;	Variables
;
;***********************************************************************


VARS		UDATA


		global	Sys_fault
Sys_fault	res	1		; System fault flags

Timer_20ms	res	1		; 20 msec timer


;***********************************************************************
;
;	Vectors
;
;***********************************************************************

STARTUP	CODE

	goto	Main		; Reset vector
	nop
	nop
	nop

	extern	Interrupt
	goto	Interrupt	; Interrupt vector


PROG	CODE

;***********************************************************************
;
;	Tables
;
;	These must be at the beginning of code
;	so they will reside in Page 0
;
;***********************************************************************

; The key output table consists of the values to be written
; to the digital pot to obtain 8 equally spaced voltage values
; with stepsize as close as possible to 12.5% of the full range.
;
; The voltage is computed assuming a 64-step 10K pot
; which is feeding a 50K load to ground.
;
;	             (N/63)*10K || 50K
;	V = ------------------------------------
;	    ((N/63)*10K || 50K) + (1 - N/63)*10K
;
; Rather than trying to solving for N, I just used trial and error
; to come up with the following values. Note the worst case percentage
; error is no larger than 0.5% from the ideal value of 12.5% * (1 to 8).
; This is well within the spa control pack's key detection tolerance.

	global	KeyoutTbl
KeyoutTbl
	addwf	PCL,f
	retlw	1	; (not used)
	retlw	8	; 12.4%
	retlw	16	; 24.5%
	retlw	25	; 37.9%
	retlw	33	; 49.9%
	retlw	41	; 62.2%
	retlw	49	; 75.2%
	retlw	56	; 87.2%
	retlw	63	; 100%


; State machine jump table

	global	StateTbl
StateTbl
	addwf	PCL,f
	extern	State0
	goto	State0
	extern	State1
	goto	State1
	extern	State2
	goto	State2
	extern	State3
	goto	State3
	extern	State4
	goto	State4
	extern	State5
	goto	State5
	extern	State6
	goto	State6
	extern	State7
	goto	State7


;***********************************************************************
;
;	Main routine
;
;***********************************************************************

Main
	call	InitHw		; Init hardware

	call	InitSw		; Init software

	movf	SSPBUF,w
	CALLX	LCDprep		; Prepare SSP port for first transfer

	SETBANK	1
	MOVE	(1<<SSPIE),PIE1	; enable SSP interrupt
	SETBANK	0
	MOVE	((1<<PEIE)|(1<<GIE)),INTCON

MainLp
	clrwdt

	CALLX	DoState		; Execute main state machine

	CALLX	DoLCD		; Do LCD processing

	CALLX	DoKeyIn		; Do keypad decoding

	CALLX	DoKeyOut	; Do key output processing

	CALLX	DoPump		; Do pump control processing

	CALLX	DoSerial	; Do serial I/O processing

	call	DoTimer		; Do timer processing

	goto	MainLp


;***********************************************************************
;
;	Timer subroutine
;
;***********************************************************************

;  Handle the 1 msec tick timer

DoTimer
	btfss	INTCON,T0IF	; Timer interrupt flag set?
	return			;  no

	bcf	INTCON,T0IF	; Clear flag

	extern	LCD_timer
	movf	LCD_timer,w	; LCD frame timer nonzero?
	SKIPZ
	decf	LCD_timer,f	;  yes, decrement it

	extern	Key_timer
	movf	Key_timer,w	; Key scan timer nonzero?
	SKIPZ
	decf	Key_timer,f	;  yes, decrement it

	extern	Ping_timer
	movf	Ping_timer,w	; Serial ping timer nonzero?
	SKIPZ
	decf	Ping_timer,f	;  yes, decrement it

	extern	Resp_timer
	movf	Resp_timer,w	; Serial comm response timer nonzero?
	SKIPZ
	decf	Resp_timer,f	;  yes, decrement it

	decfsz	Timer_20ms,f	; 20 msec interval timer expired?
	goto	No20ms		;  no
	movlw	20
	movwf	Timer_20ms	;  yes, restart it

;	Do 20 msec interval processing

	extern	Keyout_timer
	movf	Keyout_timer,w	; Key output duration timer nonzero?
	SKIPZ
	decf	Keyout_timer,f	;  yes, decrement it

	extern	State_timer
	movf	State_timer,w	; State mode timer nonzero?
	SKIPZ
	decf	State_timer,f	;  yes, decrement it

No20ms
	return


TimerInit
	movlw	20
	movwf	Timer_20ms	; Init timer count
	return

;***********************************************************************
;
;	ADC subroutine
;
;***********************************************************************

;  Get an ADC sample

	global	GetADC
GetADC
	iorlw	1<<ADON
	movwf	ADCON0		; Enable ADC and select channel

	movlw	10		; Wait 10 usec for sampling interval
ADClp1	addlw	-1
	BNE	ADClp1

	bsf	ADCON0,GO_DONE	; Start conversion
ADClp2	btfsc	ADCON0,GO_DONE	; Wait for it to complete
	goto	ADClp2		;  (about 44 usec with a 4 MHz clock)

	movf	ADRESH,w	; Get upper 8 bits of result
	return


;***********************************************************************
;
;	Initialization
;
;***********************************************************************

; Initialize hardware peripherals

InitHw
	SETBANK	2

;	Init comparator

	MOVE	0x97,CM1CON0	; Set Cmp1 to inverted C12IN3- and C1Vref+
	MOVE	0xA2,VRCON	; Set C1Vref to Vdd/12

;	Init I/O ports

	MOVE	0xF0,ANSEL	; Set RC3:0 to analog inputs
	MOVE	0x00,ANSELH

	MOVE	0x00,WPUB	; Disable all port B pullups

	MOVE	0x00,PORTA	; Init all port bits
	MOVE	0x00,PORTB
	MOVE	0x10,PORTC

	SETBANK	1

	MOVE	0xFC,TRISA	; Set RA1:0 to output
	MOVE	0x4F,TRISC	; Set RC7,5,4 to output
	MOVE	0x04,WPUA	; Enable pullup on RA2

;	Init ADC

	MOVE	0x50,ADCON1	; Set ADC clock to Fosc/16

;	Init UART

	MOVE	0x24,TXSTA	; Enable transmitter, high speed
	MOVE	0x10,BAUDCTL	; Set 8-bit baud, inverted transmit
_BAUD	equ	((4000000/16)+(BAUD/2))/BAUD
	MOVE	_BAUD-1,SPBRG	; Set 9600 baud

;	Init SSP

	MOVE	0x00,SSPSTAT

;	Init options

	clrwdt			; Clear watchdog before changing prescaler
	MOVE	0x01,OPTION_REG	; Enable pullups, Timer0 1/4 prescaler

	SETBANK	0

;	Init UART

	MOVE	0x90,RCSTA	; Enable receiver

;	Init SSP

	MOVE	0x34,SSPCON	; Enable SPI slave port with /SS

;	Init timers

	MOVE	0x3F,T1CON	; Enable Timer1 from external LP osc

	return


; Initialize software & variables

InitSw
	CALLX	LCDinit
	CALLX	KeyInit
	CALLX	PotClr
	CALLX	SerInit
	call	TimerInit
	CALLX	StateInit
	clrf	Sys_fault
	return

	end
