;***********************************************************************
;
;	Microchip 16-bit Embedded Control Design Contest
;
;	Entry # MT2268
;
;	Spa Pump Controller - Spa Interface Processor
;
;***********************************************************************
;
;	Main state processing
;
;***********************************************************************


#include	<P16f687.inc>
#include	"SpaCtl.inc"


;***********************************************************************
;
;	Variables
;
;***********************************************************************

	extern	Key_flag
	extern	Key_val
	extern	Keyout_timer
	extern	LCD_icons
	extern	Sys_fault


VARS	UDATA

		global	State
State		res	1		; System state

		global	State_timer
State_timer	res	1		; System state timer


PROG	CODE


;***********************************************************************
;
;	Main State Processing Routine
;
;***********************************************************************
;
; State handler called by main routine
;

	global	DoState
DoState
	movf	State,w		; Get current state
	andlw	7		; Make sure it's valid
	extern	StateTbl
	goto	StateTbl	; Go to jump table

;***********************************************************************
;
;	State 0
;
;	Wait for Pump 1 key to be pressed to start us off.
;	Meanwhile just echo any other incoming keypresses.
;

	global	State0
State0
	movf	Key_flag,w	; Any key pressed?
	BEQ	StDone		;  no
	xorlw	KV_pump1	; Pump 1 key pressed?
	BEQ	State0a		;  yes
	goto	StSendK		; Send original key
State0a
	incf	State,f		; Go to state 1
	movlw	KV_pump1
	goto	StSend		;  and send the pump 1 key to turn on low speed relay

;***********************************************************************
;
;	State 1
;
;	Wait for outgoing Pump 1 keypress to complete,
;	while ignoring any incoming keypresses.
;

	global	State1
State1
	movf	Keyout_timer,w	; Still sending pump 1 key?
	BNE	StDone		;  yes, wait for it to be done
	incf	State,f		; Go to state 2
	movlw	KV_none
	goto	StSend		; Send a null key for a short delay

;***********************************************************************
;
;	State 2
;
;	Send another Pump 1 keypress to turn on Pump 1 high speed relay
;	(first keypress only turned on low speed relay, which we're not using)
;

	global	State2
State2
	movf	Keyout_timer,w	; Still sending null key?
	BNE	StDone		;  yes, wait for it to be done
	incf	State,f		; Go to state 2
	movlw	KV_pump1
	goto	StSend		; Send pump 1 key again to turn on high speed relay

;***********************************************************************
;
;	State 3
;
;	Wait for outgoing Pump 1 keypress to complete.
;	Then send initialization command to pump drive unit,
;	and start a 5 second timer to let user adjust pump speed.
;

	global	State3
State3
	movf	Keyout_timer,w	; Still sending pump 1 key?
	BNE	StDone		;  yes, wait for it to be done
	CALLX	Pump1Init	; Init pump speed control
	incf	State,f		; Go to state 4
StTimer
	movlw	5000/20		; Start a 5-second mode timer
	movwf	State_timer	;  to allow pump up/down speed adjustment
	goto	StDone

;***********************************************************************
;
;	State 4
;
;	This state handles the pump speed setting mode, where the
;	UP/DOWN keys can be used to increase or decrease the speed.
;	The keys will autorepeat if held down. This state only lasts
;	until the 5 second timer runs out, after which the UP/DOWN
;	keys return to their normal function of temperature setting.
;	Pressing (but not holding) the Pump 1 key will turn it off.
;	Pressing any other key will exit this speed setting mode.
;

	global	State4
State4
	movf	Sys_fault,w	; Any fault flags set?
	BNE	StateOff	;  yes, turn pump off immediately
	movf	State_timer,w	; Mode timer still running?
	BEQ	State4t		;  no
	movf	Key_flag,w	; Any key pressed,
	andlw	0x0F		;  ignoring hold/repeat flags?
	BEQ	StDone		;  no
	xorlw	KV_up		; Up key pressed or repeated?
	BNE	State4a		;  no
	CALLX	Pump1Up		; Increase pump speed
	goto	StTimer		;  and restart mode timer
State4a	xorlw	KV_down^KV_up	; Down key pressed or repeated?
	BNE	State4b		;  no
	CALLX	Pump1Down	; Decrease pump speed
	goto	StTimer		;  and restart mode timer
State4b	xorlw	KV_pump1^KV_down; Pump 1 key pressed?
	BNE	State4c		;  no
	btfsc	Key_flag,K_HOLD	; Is it just held down from before?
	goto	StDone		;  yes, ignore it
	CALLX	Pump1Off	; Turn off pump controller
	movlw	200/20		; Start a 200 msec shutdown timer
	movwf	State_timer
	movlw	7
	movwf	State		; Go to state 7
	goto	StDone
State4c	clrf	State_timer	; Clear mode timer if any other key pressed
	incf	State,f		;  and go to state 5
	goto	StSendK		; Send the key
State4t	incf	State,f		; Go to state 5 if timer expired
	goto	StDone

;***********************************************************************
;
;	State 5
;
;	This is the normal running state. Any keypresses are just
;	echoed, except for Pump 1 which is handled by the next state.
;

	global	State5
State5
	movf	Sys_fault,w	; Any fault flags set?
	BNE	StateOff	;  yes, turn pump off immediately
	movf	Key_flag,w	; Any key pressed?
	BEQ	StDone		;  no
	xorlw	KV_pump1	; Pump 1 key pressed?
	BNE	StSendK		;  no, send the key
	incf	State,f		; Go to state 6
	goto	StDone

;***********************************************************************
;
;	State 6
;
;	This handles a Pump 1 keypress during normal operation.
;	If it's pressed and held, we return to speed setting mode
;	and restart the 5 second timer. If it's pressed and released,
;	pump 1 is turned off.
;

	global	State6
State6
	movf	Sys_fault,w	; Any fault flags set?
	BNE	StateOff	;  yes, turn pump off immediately
	movf	Key_flag,w	; Any key pressed?
	BEQ	State6a		;  no
	xorlw	KV_pump1+KV_hold ; Pump 1 key held down?
	BNE	StSendK		;  no, send the key
	movlw	4
	movwf	State		; Go back to state 4
	goto	StTimer		; Restart 5-second mode timer
State6a	movf	Key_val,w
	xorlw	KV_pump1	; Pump 1 key still pressed?
	BEQ	StDone		;  yes
	CALLX	Pump1Off	; Turn off pump controller
	movlw	200/20		; Start a 200 msec shutdown timer
	movwf	State_timer
	incf	State,f		; Go to state 7
	goto	StDone

;***********************************************************************
;
;	State 7
;
;	Wait for a short delay for pump drive unit to complete
;	its shutdown processing, then send out a Pump 1 keypress
;	to turn off the pump relay. If the LCD icons indicate the
;	pump wasn't on, then we don't do anything, which syncs
;	us up to the spa control pack's relay state if we got
;	out of sync.
;

	global	State7
State7
	movf	Sys_fault,w	; Any fault flags set?
	BNE	StateOff	;  yes, turn pump off immediately
	movf	State_timer,w	; Shutdown timer elapsed?
	BNE	StDone		;  no
StateOff
	clrf	State		; Go back to state 0
	btfss	LCD_icons,ICON_pump1 ; Pump 1 LCD icon on?
	goto	StDone		;  no, all done
	movlw	KV_pump1	; Send pump 1 key to turn off the pump relay
	goto	StSend

;***********************************************************************

StSendK	movf	Key_flag,w	; Get the current incoming keypress
StSend	CALLX	SendKey		; Send out the keypress
StDone	clrf	Key_flag	; Clear the incoming key pressed flag
	return


;***********************************************************************
;
;	Initialization
;
;***********************************************************************

	global	StateInit
StateInit
	clrf	State		; Start off at state 0
	return

	end
